@@ -1,6 +1,7 @@
+import math
+
 import pytest
 
-from jinja2._compat import text_type
 from jinja2.exceptions import UndefinedError
 from jinja2.nativetypes import NativeEnvironment
 from jinja2.nativetypes import NativeTemplate
@@ -53,7 +54,7 @@ def test_multi_expression_add(env):
 def test_loops(env):
     t = env.from_string("{% for x in value %}{{ x }}{% endfor %}")
     result = t.render(value=["a", "b", "c", "d"])
-    assert isinstance(result, text_type)
+    assert isinstance(result, str)
     assert result == "abcd"
 
 
@@ -111,7 +112,7 @@ def test_constant_dunder_to_string(env):
 def test_string_literal_var(env):
     t = env.from_string("[{{ 'all' }}]")
     result = t.render()
-    assert isinstance(result, text_type)
+    assert isinstance(result, str)
     assert result == "[all]"
 
 
@@ -140,9 +141,22 @@ def test_no_intermediate_eval(env):
     assert isinstance(result, float)
     # If intermediate eval happened, 0.000 would render 0.0, then 7
     # would be appended, resulting in 0.07.
-    assert result < 0.007  # TODO use math.isclose in Python 3
+    assert math.isclose(result, 0.0007)
 
 
 def test_spontaneous_env():
     t = NativeTemplate("{{ true }}")
     assert isinstance(t.environment, NativeEnvironment)
+
+
+def test_leading_spaces(env):
+    t = env.from_string(" {{ True }}")
+    result = t.render()
+    assert result == " True"
+
+
+def test_macro(env):
+    t = env.from_string("{%- macro x() -%}{{- [1,2] -}}{%- endmacro -%}{{- x()[1] -}}")
+    result = t.render()
+    assert result == 2
+    assert isinstance(result, int)
